/*
Gecko Deck 1.0
(c) 2021 Gecko Tribe, LLC
by Antone Roundy
https://streamer.geckotribe.com/geckodeck/
You may freely use Gecko Deck for private and commercial purposesk, and may modify it for you own use.
Redistribution, except to people you know personally, is prohibited.
*/
"use strict";

var fs = require('fs'),
	http = require('http');
const { keyboard, Key } = require("@nut-tree/nut-js");

var keyDelay=25;
var GeckoDeckPort=8080;

function sleep(ms) {
	return new Promise((resolve) => {
		setTimeout(resolve, ms);
	});
}  

const DoKD = async (k) => {
	var i,ks=[];
	if (k.length==1) ks=k;
	else for (i=0;i<k.length;i++) ks[i]=keymap[k[i]];
	switch(k.length) {
	case 1: await keyboard.pressKey(ks); break;
	case 2: await keyboard.pressKey(ks[0],ks[1]); break;
	case 3: await keyboard.pressKey(ks[0],ks[1],ks[2]); break;
	case 4: await keyboard.pressKey(ks[0],ks[1],ks[2],ks[3]); break;
	case 5: await keyboard.pressKey(ks[0],ks[1],ks[2],ks[3],ks[4]); break;
	}	
};
const DoKU = async (k) => {
	var i,ks=[];
	if (k.length==1) ks=k;
	else for (i=0;i<k.length;i++) ks[i]=keymap[k[i]];
	switch(k.length) {
	case 1: await keyboard.releaseKey(ks); break;
	case 2: await keyboard.releaseKey(ks[0],ks[1]); break;
	case 3: await keyboard.releaseKey(ks[0],ks[1],ks[2]); break;
	case 4: await keyboard.releaseKey(ks[0],ks[1],ks[2],ks[3]); break;
	case 5: await keyboard.releaseKey(ks[0],ks[1],ks[2],ks[3],ks[4]); break;
	}	
};

const DoGeckoDeckCmd = async (args) => {
	var i;
	for (i=0;i<args.length;i++) {
		switch(args[i].c) {
		case 'kd': await DoKD(args[i].k); break;
		case 'ku': await DoKU(args[i].k); break;
		case 'kdu':
			await DoKD(args[i].k);
			await sleep((typeof(args[i].t)==='undefined')?keyDelay:args[i].t);
			await DoKU(args[i].k);
			break;
		case 't': await keyboard.type(args[i].k); break;
		case 'w': await sleep((typeof(args[i].t)==='undefined')?keyDelay:args[i].t); break;
		case 'skd': keyDelay=args[i].t;
		//!!! any way to locate and switch to a window? (be sure we can restore the current window afterward)
		}
	}
};

function do404(res,err) {
	res.writeHead(404);
	res.write((typeof(err)==='undefined')?'Not found.':err);
	res.end();
}

http.createServer(function (req, res) {
	var urlp,fnp,lo,i;
	urlp=(req.url=='/')?['layout','home.json']:req.url.substr(1).split('/');
	if (urlp.length>1) urlp[1]=urlp[1].replace(/[^-_a-zA-Z0-9.:]/g,'').replace(/[.]*:/g,'/');
	switch (urlp[0]) {
	case 'layout':
		try {
			if (fs.existsSync(__dirname+'/layout/'+urlp[1])) {
				fnp=urlp[1].split('.');
				res.setHeader('Content-Type','text/html');
				fs.readFile(__dirname+'/layout/'+urlp[1], function (err,data) {
					if (err) do404(res,JSON.stringify(err));
					else {
						lo=JSON.parse(data.toString().replace(/\/\/.*$/gm,''));
						if (typeof(lo.keyDelay)!=='undefined') keyDelay=lo.keyDelay;
						res.write("<!DOCTYPE HTML>\n"
							+'<html><head>'
							+'<meta name="viewport" content="width=device-width, initial-scale=1,maximum-scale=1"/>'
							+"<title>GeckoDeck Layout</title>"
							+'<script type="text/javascript" src="/js/geckodeck.js"></script>'
							+'<link rel="manifest" href="/manifest.json">'
							//+'<meta name="mobile-web-app-capable" content="yes">'
							);
						if (typeof(lo.css)!=='undefined')
							for (i=0;i<lo.css.length;i++) res.write('<link rel="stylesheet" type="text/css" href="/css/'+lo.css[i]+'"/>');
						res.write('</head><body><div id="all">');
						for (i=0;i<lo.buttons.length;i++) {
							res.write('<a href="'
									+((typeof(lo.buttons[i].layout)!=='undefined')?('/layout/'+lo.buttons[i].layout):'#')
										+'"'
									+((typeof(lo.buttons[i].actions)!=='undefined')?(' onClick="return ClickButton(\''+urlp[1]+'\',\''+i+'\');"'):
										((typeof(lo.buttons[i].layout)!=='undefined')?'':' onClick="return false;"'))
									+((typeof(lo.buttons[i].img)==='undefined')?' class="noimg"':'')
									+'>'
								+((typeof(lo.buttons[i].img)!=='undefined')?('<img src="/img/'+lo.buttons[i].img+'"/>'):'')
								+((typeof(lo.buttons[i].label)!=='undefined')?('<span>'+lo.buttons[i].label+'</span>'):'')
								+'</a>');
						}
						res.write('</div></body></html>');
						res.end();
					}
				});
			}
		} catch(err) {
			do404(res,JSON.stringify(err));
		}
		break;
	case 'img':
		try {
			if (fs.existsSync(__dirname+'/img/'+urlp[1])) {
				fnp=urlp[1].split('.');
				var mime=fnp[fnp.length-1];
				if (mime=='svg') mime='svg+xml';
				res.setHeader('Content-Type','image/'+mime);
				fs.readFile(__dirname+'/img/'+urlp[1], function (err,data) {
					if (err) do404(res,JSON.stringify(err));
					else {
						res.write(data);
						res.end();
					}
				});
			} else do404(res);
		} catch(err) {
			do404(res,JSON.stringify(err));
		}
		break;
	case 'js':
		try {
			if (fs.existsSync(__dirname+'/js/'+urlp[1])) {
				res.setHeader('Content-Type','text/javascript');
				fs.readFile(__dirname+'/js/'+urlp[1], function (err,data) {
					if (err) do404(res,JSON.stringify(err));
					else {
						res.write(data);
						res.end();
					}
				});
			} else do404(res);
		} catch(err) {
			do404(res,JSON.stringify(err));
		}
		break;
	case 'css':
		try {
			if (fs.existsSync(__dirname+'/css/'+urlp[1])) {
				res.setHeader('Content-Type','text/css');
				fs.readFile(__dirname+'/css/'+urlp[1], function (err,data) {
					if (err) do404(res,JSON.stringify(err));
					else {
						res.write(data);
						res.end();
					}
				});
			} else do404(res);
		} catch(err) {
			do404(res,JSON.stringify(err));
		}
		break;
	case 'cmd':
		try {
			keyboard.config.autoDelayMs=25;
			if (fs.existsSync(__dirname+'/layout/'+urlp[1])) {
				fnp=urlp[1].split('.');
				fs.readFile(__dirname+'/layout/'+urlp[1], function (err,data) {
					if (err) do404(res,JSON.stringify(err));
					else {
						res.setHeader('Content-Type','text/plain');
						lo=JSON.parse(data.toString().replace(/\/\/.*$/gm,''));
						if ((typeof(lo.buttons[urlp[2]])!=='undefined')&&(typeof(lo.buttons[urlp[2]].actions)!=='undefined')) {
							DoGeckoDeckCmd(lo.buttons[urlp[2]].actions);
							res.write('1');
						} else res.write('0');
						res.end();
					}
				});
			}
		} catch(err) {
			do404(res,JSON.stringify(err));
		}
		break;
	case 'manifest.json':
		res.write('{'
			+'"short_name":"Gecko Deck"'
			+',"name":"Gecko Deck"'
			+',"start_url":"/"'
			+',"display":"fullscreen"'
			+'}');
		res.end();
		break;
	default: do404(res);
	}
}).listen(GeckoDeckPort);

// https://nut-tree.github.io/nut.js/enums/key.html
const keymap={
"Space":Key.Space
,"Escape":Key.Escape
,"Tab":Key.Tab

,"LeftAlt":Key.LeftAlt
,"LeftControl":Key.LeftControl
,"RightAlt":Key.RightAlt
,"RightControl":Key.RightControl

,"LeftShift":Key.LeftShift
,"LeftSuper":Key.LeftSuper
,"RightShift":Key.RightShift
,"RightSuper":Key.RightSuper

,"F1":Key.F1
,"F2":Key.F2
,"F3":Key.F3
,"F4":Key.F4
,"F5":Key.F5
,"F6":Key.F6
,"F7":Key.F7
,"F8":Key.F8
,"F9":Key.F9
,"F10":Key.F10
,"F11":Key.F11
,"F12":Key.F12

,"Num0":Key.Num0
,"Num1":Key.Num1
,"Num2":Key.Num2
,"Num3":Key.Num3
,"Num4":Key.Num4
,"Num5":Key.Num5
,"Num6":Key.Num6
,"Num7":Key.Num7
,"Num8":Key.Num8
,"Num9":Key.Num9

,"A":Key.A
,"B":Key.B
,"C":Key.C
,"D":Key.D
,"E":Key.E
,"F":Key.F
,"G":Key.G
,"H":Key.H
,"I":Key.I
,"J":Key.J
,"K":Key.K
,"L":Key.L
,"M":Key.M
,"N":Key.N
,"O":Key.O
,"P":Key.P
,"Q":Key.Q
,"R":Key.R
,"S":Key.S
,"T":Key.T
,"U":Key.U
,"V":Key.V
,"W":Key.W
,"X":Key.X
,"Y":Key.Y
,"Z":Key.Z

,"Grave":Key.Grave
,"Minus":Key.Minus
,"Equal":Key.Equal
,"Backspace":Key.Backspace
,"LeftBracket":Key.LeftBracket
,"RightBracket":Key.RightBracket
,"Backslash":Key.Backslash
,"Semicolon":Key.Semicolon
,"Quote":Key.Quote
,"Return":Key.Return
,"Comma":Key.Comma
,"Period":Key.Period
,"Slash":Key.Slash

,"Left":Key.Left
,"Up":Key.Up
,"Right":Key.Right
,"Down":Key.Down

,"Print":Key.Print
,"Pause":Key.Pause
,"Insert":Key.Insert
,"Delete":Key.Delete
,"Home":Key.Home
,"End":Key.End
,"PageUp":Key.PageUp
,"PageDown":Key.PageDown

,"Add":Key.Add
,"Subtract":Key.Subtract
,"Multiply":Key.Multiply
,"Divide":Key.Divide
,"Decimal":Key.Decimal
,"Enter":Key.Enter

,"NumPad0":Key.NumPad0
,"NumPad1":Key.NumPad1
,"NumPad2":Key.NumPad2
,"NumPad3":Key.NumPad3
,"NumPad4":Key.NumPad4
,"NumPad5":Key.NumPad5
,"NumPad6":Key.NumPad6
,"NumPad7":Key.NumPad7
,"NumPad8":Key.NumPad8
,"NumPad9":Key.NumPad9

,"CapsLock":Key.CapsLock
,"ScrollLock":Key.ScrollLock
,"NumLock":Key.NumLock
};
